history row .code-label {
  font-weight: bold;
}

history row {
  padding: 18px;
}

history .content.small > row:first-child {
  border-top-width: 0;
}

history > row:first-child {
  margin-top: 12px;
}

history > row:last-child {
  margin-bottom: 12px;
}

history > row {
  margin-left: 12px;
  margin-right: 12px;
}

textview {
  padding: 6px;
}

toolbarview scrolledwindow.card undershoot.top,
toolbarview scrolledwindow.card undershoot.bottom {
  box-shadow: none;
  background: none;
}

.camera-view {
  background-color: black;
}
